/*
 * Copyright (C) 2012 Technische Universitaet Muenchen
 * This file is part of the Sierpinski project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpinski
 *
 *  Created on: 31. October 2012
 *      Author: Michael Müller <muellmic@in.tum.de>
 */

#ifndef CCUBE_TO_SPHERE_PROJECTIONS_HPP_
#define CCUBE_TO_SPHERE_PROJECTIONS_HPP_

#include <cmath>



/**
 * This class contains helper methods to project
 * 2D world coordinates to 2D coordinates on a cube face
 * and 3D coordinates on a unit sphere.
 */
class CCube_To_Sphere_Projection
{
	typedef CONFIG_DEFAULT_FLOATING_POINT_TYPE T;
public:

	/**
	 * Faces of the cube which is projected to the cubed sphere
	 */
	enum EFaceType {
		E_ERROR = -1,
		E_FACE_FRONT = 0,
		E_FACE_RIGHT = 1,
		E_FACE_BACK = 2,
		E_FACE_LEFT = 3,
		E_FACE_TOP = 4,
		E_FACE_BOTTOM = 5
	};



public:
	/**
	 * This method identifies the cube face of a triangle from its 2D position,
	 * assuming that the cube map lies in a 8x8 world, linearly scaled in x and
	 * y direction and centered at (0,0).
	 *
	 *
	 *
	 *           (-2,3)      (0,3)
	 *              +----------+
	 *              |          |
	 *              |          |
	 *              |   TOP    |
	 *(-4,1)        |          |        (2,1)      (4,1)
	 *   +----------+----------+----------+----------+
	 *   |          |          |          |          |
	 *   |          |          |          |          |
	 *   |   LEFT   |  FRONT   |  RIGHT   |   BACK   |
	 *   |          |          |          |          |
	 *   +----------+----------+----------+----------+
	 *(-4,-1)       |          |        (2,-1)     (4,-1)
	 *              |          |
	 *              |  BOTTOM  |
	 *              |          |
	 *              +----------+
	 *           (-2,-3)     (0,-3)
	 *
	 */
	static inline EFaceType getFaceIdOf2DCubeCoordinate(
		T i_x,	T i_y
	) {
		// compare sum of x- or y-coordinates to 3 times the interval boundaries

		// horizontal band
		if(	(i_y >= (T)-1.0) && (i_y <= (T)1.0)	)		// y between -1 and 1
		{
			if(i_x < (T)-4.0){		// x < -4	ERROR
				return E_ERROR;
			} else if(i_x < (T)-2.0) {	// x between -4 and -2
				return E_FACE_LEFT;
			} else if(i_x < (T)0.0) {	// x between -2 and 0
				return E_FACE_FRONT;
			} else if(i_x < (T)2.0) {	// x between 0 and 2
				return E_FACE_RIGHT;
			} else if(i_x <= (T)4.0) {	// x between 2 and 4
				return E_FACE_BACK;
			}
		}

		// vertical band
		if(	(i_x >= (T)-2.0) && (i_x <= (T)0.0)	)		// x between -2 and 0
		{
			if(	i_y < (T)-3.0) {		// y < -3	ERROR
				return E_ERROR;
			} else if(i_y < (T)-1.0) {	// y between -3 and -1
				return E_FACE_BOTTOM;
			} else if(i_y < (T)1.0) {	// y between -1 and 1
				return E_FACE_FRONT;
			} else if(i_y <= (T)3.0) {	// y between 1 and 3
				return E_FACE_TOP;
			}
		}

#if DEBUG
		std::cerr << "WARNING: Invalid vertex in method getFaceIdOf2DCubeCoordinate (" << i_x << " " << i_y << ")" << std::endl;
#endif
		return E_ERROR;
	}




	/**
	 * This method projects the 2D world coordinates of a triangle to the
	 * 2D coordinates of the corresponding cube face, assuming that the
	 * cube map lies in a 8x8 world, linearly scaled in x and y direction
	 * and centered at (0,0).
	 *
	 * INPUT REFERENCE SYSTEM:
	 *
	 *           (-2,3)      (0,3)
	 *              +----------+
	 *              |          |
	 *              |          |
	 *              |   TOP    |
	 *(-4,1)        |          |        (2,1)      (4,1)
	 *   +----------+----------+----------+----------+
	 *   |          |          |          |          |
	 *   |          |          |          |          |
	 *   |   LEFT   |  FRONT   |  RIGHT   |   BACK   |
	 *   |          |          |          |          |
	 *   +----------+----------+----------+----------+
	 *(-4,-1)       |          |        (2,-1)     (4,-1)
	 *              |          |
	 *              |  BOTTOM  |
	 *              |          |
	 *              +----------+
	 *           (-2,-3)     (0,-3)
	 *
	 *
	 *
	 * OUTPUT REFERENCE SYSTEM:
	 *
	 * (-1,1)      (1,1)
	 *    +----------+
	 *    |          |
	 *    |          |
	 *    |   FACE   |
	 *    |          |
	 *    +----------+
	 * (-1,-1)     (1,-1)
	 */
	template <typename T>
	inline static void project2DToCubeFace(
			CCube_To_Sphere_Projection::EFaceType i_face,
			T *io_vertex_x,	T *io_vertex_y
	)	{

		// translation to face
		switch(i_face){ // no change in y for LEFT, FRONT, RIGHT and BACK face
		case CCube_To_Sphere_Projection::E_FACE_LEFT:
			*io_vertex_x = *io_vertex_x + (T)3.0;
			break;

		case CCube_To_Sphere_Projection::E_FACE_FRONT:
			*io_vertex_x = *io_vertex_x + (T)1.0;
			break;

		case CCube_To_Sphere_Projection::E_FACE_RIGHT:
			*io_vertex_x = *io_vertex_x - (T)1.0;
			break;

		case CCube_To_Sphere_Projection::E_FACE_BACK:
			*io_vertex_x = *io_vertex_x - (T)3.0;
			break;

		case CCube_To_Sphere_Projection::E_FACE_TOP:
			*io_vertex_x = *io_vertex_x + (T)1.0;
			*io_vertex_y = *io_vertex_y - (T)2.0;
			break;

		case CCube_To_Sphere_Projection::E_FACE_BOTTOM:
			*io_vertex_x = *io_vertex_x + (T)1.0;
			*io_vertex_y = *io_vertex_y + (T)2.0;
			break;

		default:
		case CCube_To_Sphere_Projection::E_ERROR:
			std::cerr << "project2DToCubeFace: invalid face" << std::endl;
//			assert(false);
			break;
		}
	}


	template <typename T>
	inline static void project2DToCubeFace(
			CCube_To_Sphere_Projection::EFaceType i_face,
			T *io_vertex_left_x,	T *io_vertex_left_y,
			T *io_vertex_right_x,	T *io_vertex_right_y,
			T *io_vertex_top_x,		T *io_vertex_top_y
	)	{

		// translation to face
		switch(i_face)
		{ // no change in y for LEFT, FRONT, RIGHT and BACK face
		case CCube_To_Sphere_Projection::E_FACE_LEFT:
			*io_vertex_left_x = *io_vertex_left_x + (T)3.0;
			*io_vertex_right_x = *io_vertex_right_x + (T)3.0;
			*io_vertex_top_x = *io_vertex_top_x + (T)3.0;
			break;

		case CCube_To_Sphere_Projection::E_FACE_FRONT:
			*io_vertex_left_x = *io_vertex_left_x + (T)1.0;
			*io_vertex_right_x = *io_vertex_right_x + (T)1.0;
			*io_vertex_top_x = *io_vertex_top_x + (T)1.0;
			break;

		case CCube_To_Sphere_Projection::E_FACE_RIGHT:
			*io_vertex_left_x = *io_vertex_left_x - (T)1.0;
			*io_vertex_right_x = *io_vertex_right_x - (T)1.0;
			*io_vertex_top_x = *io_vertex_top_x - (T)1.0;
			break;

		case CCube_To_Sphere_Projection::E_FACE_BACK:
			*io_vertex_left_x = *io_vertex_left_x - (T)3.0;
			*io_vertex_right_x = *io_vertex_right_x - (T)3.0;
			*io_vertex_top_x = *io_vertex_top_x - (T)3.0;
			break;

		case CCube_To_Sphere_Projection::E_FACE_TOP:
			*io_vertex_left_x = *io_vertex_left_x + (T)1.0;
			*io_vertex_left_y = *io_vertex_left_y - (T)2.0;
			*io_vertex_right_x = *io_vertex_right_x + (T)1.0;
			*io_vertex_right_y = *io_vertex_right_y - (T)2.0;
			*io_vertex_top_x = *io_vertex_top_x + (T)1.0;
			*io_vertex_top_y = *io_vertex_top_y - (T)2.0;
			break;

		case CCube_To_Sphere_Projection::E_FACE_BOTTOM:
			*io_vertex_left_x = *io_vertex_left_x + (T)1.0;
			*io_vertex_left_y = *io_vertex_left_y + (T)2.0;
			*io_vertex_right_x = *io_vertex_right_x + (T)1.0;
			*io_vertex_right_y = *io_vertex_right_y + (T)2.0;
			*io_vertex_top_x = *io_vertex_top_x + (T)1.0;
			*io_vertex_top_y = *io_vertex_top_y + (T)2.0;
			break;

		default:
		case CCube_To_Sphere_Projection::E_ERROR:
			std::cerr << "project2DToCubeFace: invalid face id " << i_face << std::endl;
//			assert(false);
			break;
		}
	}



	template <typename T>
	inline static void project2DToLonLat(
			T i_vertex_x, 	// face based x-coordinate
			T i_vertex_y, 	// face based y-coordinate,

			T *o_lon,	// longitude
			T *o_lat	// latitude
	)
	{
		T lon, lat;
		T x = i_vertex_x;
		T y = i_vertex_y;

#if 0
		T o_x, o_y, o_z;
		CCube_To_Sphere_Projection::compute3DFrom2DCoordinate(
				x, y,
				o_x, o_y, o_z
		);

		// equidistant
		lat = std::atan(o_z/std::sqrt(o_x*o_x+o_y*o_y));
		lon = std::atan(std::abs(o_x/o_y));

		lon = (lon/M_PI_2)*(T)90.0;
		lat = (lat/M_PI_2)*(T)90.0;

		if (o_y < 0)
			lon = 180 - lon;
		if (o_x < 0)
			lon = -lon;

		lon = 180 - lon;

#else
		T o_x, o_y, o_z;
		CCube_To_Sphere_Projection::compute3DFrom2DCoordinate(
				x, y,
				o_x, o_y, o_z
		);

		CCube_To_Sphere_Projection::EFaceType face_id = CCube_To_Sphere_Projection::getFaceIdOf2DCubeCoordinate(x, y);
		CCube_To_Sphere_Projection::project2DToCubeFace(face_id, &x, &y);

		lon = std::atan(x);
		lat = std::atan(y*std::cos(lon));

		lon = (lon/M_PI_2)*(T)90.0+(T)45.0;
		lat = (lat/M_PI_2)*(T)90.0;

		lon += 90*(int)face_id;
		lon += 270;

		if (lon > 360)
			lon -= 360;
/*
		if (o_y < -o_x)
		{
			if (o_x < o_y)
				lon = 270 + lon;
			else
				lon = 180 + lon;
		}
		else
		{
			if (x > y)
				lon = 90 + lon;
		}
*/
#endif
		*o_lon = lon;
		*o_lat = lat;
	}


	/**
	 * This method projects face based coordinates of a unit-sphere-bounding cube onto
	 * the unit sphere using central projection.
	 * The target coordinates are 3D sphere coordinates.
	 *
	 *
	 *
	 * INPUT REFERENCE SYSTEM:
	 *
	 * (-1,1)      (1,1)
	 *    +----------+
	 *    |          |
	 *    |          |
	 *    |   FACE   |
	 *    |          |
	 *    +----------+
	 * (-1,-1)     (1,-1)
	 *
	 *
	 * OUTPUT COODINATE SYSTEM:
	 *
	 *
	 *     z ^
	 *       |
	 *       |
	 *       |  /
	 *       | /
	 *       |/
	 *  ----   ---------->
	 *      /| 			y
	 *     / |
	 *    /  |
	 *   /
	 *  v x
	 */
	template <typename T>
	inline static void projectToSphere(
			EFaceType face,
			T i_vertex_x, 	// face based x-coordinate
			T i_vertex_y, 	// face based y-coordinate,

			T *o_sphere_x,	// 3D x-coordinate on sphere
			T *o_sphere_y,	// 3D y-coordinate on sphere
			T *o_sphere_z	// 3D z-coordinate on sphere
	)
	{

		T x,y,z;
		// transform local 2D coordinates to global 3D coordinates
		switch(face){
		case E_FACE_LEFT:
			x = i_vertex_x;
			y = (T)-1.0;
			z = i_vertex_y;
			break;

		case E_FACE_FRONT:
			x = (T)1.0;
			y = i_vertex_x;
			z = i_vertex_y;
			break;

		case E_FACE_RIGHT:
			x = -i_vertex_x;
			y = (T)1.0;
			z = i_vertex_y;
			break;

		case E_FACE_BACK:
			x = (T)-1.0;
			y = -i_vertex_x;
			z = i_vertex_y;
			break;

		case E_FACE_TOP:
			x = -i_vertex_y;
			y = i_vertex_x;
			z = (T)1.0;
			break;

		case E_FACE_BOTTOM:
			x = i_vertex_y;
			y = i_vertex_x;
			z = (T)-1.0;
			break;

		default:
			x=y=z=(T)-1;
			break;
		}

		// get vector length
		T d;
		d = std::sqrt(y*y + z*z + x*x);

		// normalize vector to the size of 1 (unit sphere)
		*o_sphere_x = (T)x / d;
		*o_sphere_y = (T)y / d;
		*o_sphere_z = (T)z / d;

	}

	/**
	 * This method projects 3D sphere coordinates to a bounding cube using central projection.
	 * The target coordinates are local coordinates on the corresponding cube face.
	 *
	 * INPUT COODINATE SYSTEM:
	 *
	 *     z ^
	 *       |
	 *       |
	 *       |  /
	 *       | /
	 *       |/
	 *  ----   ---------->
	 *      /| 			y
	 *     / |
	 *    /  |
	 *   /
	 *  v x
	 *
	 *
	 * OUTPUT REFERENCE SYSTEM:
	 *
	 * (-1,1)      (1,1)
	 *    +----------+
	 *    |          |
	 *    |          |
	 *    |   FACE   |
	 *    |          |
	 *    +----------+
	 * (-1,-1)     (1,-1)
	 */
	template <typename T>
	inline static void projectFromSphereToCube(
			T *o_vertex_x, 	// face based x-coordinate
			T *o_vertex_y, 	// face based y-coordinate,

			T i_sphere_x,	// 3D x-coordinate on sphere
			T i_sphere_y,	// 3D y-coordinate on sphere
			T i_sphere_z	// 3D z-coordinate on sphere
	)
	{

		T x,y,z;

		T d;
		EFaceType face;

		/*
		 * Get the face of the cube and largest component of the position vector.
		 *
		 * The given spherical coordinate will be mapped onto a face of the bounding cube.
		 * The face is defined by the largest component of the spherical coordinate.
		 * The sign of the largest component decides between two opposed faces.
		 */
		if(abs(i_sphere_x) >= abs(i_sphere_y) && abs(i_sphere_x) >= abs(i_sphere_z)){
			if(i_sphere_x >= 0 ){
				d = i_sphere_x;
				face = E_FACE_FRONT;

			}else{
				d = -i_sphere_x;
				face = E_FACE_BACK;
			}
		}else if(abs(i_sphere_y) >= abs(i_sphere_x) && abs(i_sphere_y) >= abs(i_sphere_z)){
			if(i_sphere_y >= 0){
				d = i_sphere_y;
				face = E_FACE_RIGHT;
			}else{
				d = -i_sphere_y;
				face = E_FACE_LEFT;
			}
		}else if(abs(i_sphere_z) >= abs(i_sphere_x) && abs(i_sphere_z) >= abs(i_sphere_y)){
			if(i_sphere_z > 0){
				d = i_sphere_z;
				face = E_FACE_TOP;
			}else {
				d = -i_sphere_z;
				face = E_FACE_BOTTOM;
			}
		}else{
			assert(false);
		}

		// normalize x,y and z, so that the largest component is 1
		x = (T)i_sphere_x / d;
		y = (T)i_sphere_y / d;
		z = (T)i_sphere_z / d;

		// turn coordinate system to match with the local coordinates
		switch(face){
		case E_FACE_LEFT:
			*o_vertex_x = x;
			//y = (T)-1.0;
			*o_vertex_y = z;
			break;

		case E_FACE_FRONT:
			//x = (T)1.0;
			*o_vertex_x = y;
			*o_vertex_y = z;
			break;

		case E_FACE_RIGHT:
			*o_vertex_x = -x;
			//y = (T)1.0;
			*o_vertex_y = z;
			break;

		case E_FACE_BACK:
			//x = (T)-1.0;
			*o_vertex_x = -y;
			*o_vertex_y = z;
			break;

		case E_FACE_TOP:
			*o_vertex_y = -x;
			*o_vertex_x = y;
			//z = (T)1.0;
			break;

		case E_FACE_BOTTOM:
			*o_vertex_y = x;
			*o_vertex_x = y;
			//z = (T)-1.0;
			break;

		default:
			assert(false);
			x=y=z=(T)-1;
			break;
		}
	}


	/**
	 * This method projects 3D sphere coordinates to the 2D map of the bounding cube
	 * using central projection. It is assumed that the cube map lies in an 8x8 world,
	 * linearly scaled in x and y direction and centered at (0,0).
	 *
	 * INPUT COODINATE SYSTEM:
	 *
	 *
	 *     z ^
	 *       |
	 *       |
	 *       |  /
	 *       | /
	 *       |/
	 *  ----   ---------->
	 *      /| 			y
	 *     / |
	 *    /  |
	 *   /
	 *  v x
	 *
	 *
	 * OUTPUT REFERENCE SYSTEM:
	 *
	 *           (-2,3)      (0,3)
	 *              +----------+
	 *              |          |
	 *              |          |
	 *              |   TOP    |
	 *(-4,1)        |          |        (2,1)      (4,1)
	 *   +----------+----------+----------+----------+
	 *   |          |          |          |          |
	 *   |          |          |          |          |
	 *   |   LEFT   |  FRONT   |  RIGHT   |   BACK   |
	 *   |          |          |          |          |
	 *   +----------+----------+----------+----------+
	 *(-4,-1)       |          |        (2,-1)     (4,-1)
	 *              |          |
	 *              |  BOTTOM  |
	 *              |          |
	 *              +----------+
	 *           (-2,-3)     (0,-3)
	 *
	 */
	template <typename T>
	inline static void projectFromSphereToWorldSpace(
		T i_sphere_x,	// 3D x-coordinate on sphere
		T i_sphere_y,	// 3D y-coordinate on sphere
		T i_sphere_z,	// 3D z-coordinate on sphere

		T *o_vertex_x, 	// world space x-coordinate
		T *o_vertex_y 	// world space y-coordinate
	)
	{
		T x,y,z;

		T d;
		EFaceType face;

		/*
		 * Get the face of the cube and largest component of the position vector.
		 *
		 * The given spherical coordinate will be mapped onto a face of the bounding cube.
		 * The face is defined by the largest component of the spherical coordinate.
		 * The sign of the largest component decides between two opposed faces.
		 */
		if(std::abs(i_sphere_x) >= std::abs(i_sphere_y) && std::abs(i_sphere_x) >= std::abs(i_sphere_z))
		{
			if (i_sphere_x >= 0 )
			{
				d = i_sphere_x;
				face = E_FACE_FRONT;
			}
			else
			{
				d = -i_sphere_x;
				face = E_FACE_BACK;
			}
		}
		else if (std::abs(i_sphere_y) >= std::abs(i_sphere_x) && std::abs(i_sphere_y) >= std::abs(i_sphere_z))
		{
			if(i_sphere_y >= 0)
			{
				d = i_sphere_y;
				face = E_FACE_RIGHT;
			}
			else
			{
				d = -i_sphere_y;
				face = E_FACE_LEFT;
			}
		}
		else if (std::abs(i_sphere_z) >= std::abs(i_sphere_x) && std::abs(i_sphere_z) >= std::abs(i_sphere_y))
		{
			if(i_sphere_z > 0)
			{
				d = i_sphere_z;
				face = E_FACE_TOP;
			}
			else
			{
				d = -i_sphere_z;
				face = E_FACE_BOTTOM;
			}
		}
		else
		{
			assert(false);
		}

		// normalize x,y and z, so that the largest component is 1
		x = (T)i_sphere_x / d;
		y = (T)i_sphere_y / d;
		z = (T)i_sphere_z / d;

		// turn and translate coordinate system to match with the world space coordinates
		switch(face)
		{
		case E_FACE_LEFT:
			*o_vertex_x = x-3;
			//y = (T)-1.0;
			*o_vertex_y = z;
			break;

		case E_FACE_FRONT:
			//x = (T)1.0;
			*o_vertex_x = y-1;
			*o_vertex_y = z;
			break;

		case E_FACE_RIGHT:
			*o_vertex_x = 1-x;
			//y = (T)1.0;
			*o_vertex_y = z;
			break;

		case E_FACE_BACK:
			//x = (T)-1.0;
			*o_vertex_x = 3-y;
			*o_vertex_y = z;
			break;

		case E_FACE_TOP:
			*o_vertex_y = 2-x;
			*o_vertex_x = y-1;
			//z = (T)1.0;
			break;

		case E_FACE_BOTTOM:
			*o_vertex_y = x-2;
			*o_vertex_x = y-1;
			//z = (T)-1.0;
			break;

		default:
			assert(false);
			x=y=z=(T)-1;
			break;
		}
	}


	template <typename T>
	static inline bool isValidCoordinate(
			T i_x,
			T i_y
	) {

		// horizontal band
		if(	(i_y >= (T)-1.0) && (i_y <= (T)1.0)	)		// y between -1 and 1
		{
			if(i_x < (T)-4.0){		// x < -4	ERROR
				return false;
			} else if(i_x < (T)-2.0) {	// x between -4 and -2
				return true;
			} else if(i_x < (T)0.0) {	// x between -2 and 0
				return true;
			} else if(i_x < (T)2.0) {	// x between 0 and 2
				return true;
			} else if(i_x <= (T)4.0) {	// x between 2 and 4
				return true;
			}
		}

		// vertical band
		if(	(i_x >= (T)-2.0) && (i_x <= (T)0.0)	)		// x between -2 and 0
		{
			if(	i_y < (T)-3.0) {		// y < -3	ERROR
				return false;
			} else if(i_y < (T)-1.0) {	// y between -3 and -1
				return true;
			} else if(i_y < (T)1.0) {	// y between -1 and 1
				return true;
			} else if(i_y <= (T)3.0) {	// y between 1 and 3
				return true;
			}
		}

		return false;
	}


	/**
	 * compute the 3D sphere coordinate based on 2D coordinate
	 */
	template <typename T>
	static inline bool compute3DFrom2DCoordinate(
			T i_x,
			T i_y,

			T &o_x,
			T &o_y,
			T &o_z
	) {
		CCube_To_Sphere_Projection::EFaceType face_point = CCube_To_Sphere_Projection::getFaceIdOf2DCubeCoordinate(i_x, i_y);

		if (face_point == E_ERROR)
			return false;

		CCube_To_Sphere_Projection::project2DToCubeFace(
				face_point,
				&i_x,&i_y
			);

		CCube_To_Sphere_Projection::projectToSphere(
				face_point,
				i_x,
				i_y,
				&o_x,
				&o_y,
				&o_z
			);

		return true;
	}
};

#endif
