! @author Alexander Breuer (breuera AT in.tum.de, http://www5.in.tum.de/wiki/index.php/Dipl.-Math._Alexander_Breuer)
! @author David George
!
! @section DESCRIPTION
! Binds GeoClaws Fortran routines to C.

! number of f-waves in the Problem (2 shock linearization = 2, augmented = 3)

module c_bind_riemannsolvers_alex
  implicit none

  contains

  ! C-interface to the augmented Riemann solver for the 1D shallow water equations.
  !   Remark: This routine is based on the respective Fortan interface (rpn2.f90)
  !           Wave-limiters are not implemented.

  subroutine c_bind_riemann_aug_JCP( maxiter,&
                                     i_variablesLeft,  i_variablesRight,&
                                     drytol, g,&
                                     o_eL, o_eR,&
                                     sw &
#if CONFIG_TSUNAMI_AUGMENTED_RIEMANN_EIGEN_COEFFICIENTS
                                     , o_eigenCoefficients &
#endif
                                     ) bind(c, name='c_bind_geoclaw_riemann_aug_JCP')
  !variable declaration
    !input
    integer                                   :: maxiter          !< maximum number of iterations over the Riemann problem.

    double precision, dimension(3)            :: i_variablesLeft,&  !< array containing variables of the left cell: $(h_l, hu_l, b_l)^T$
                                                 i_variablesRight   !< array containing variables of the right cell: $(h_r, hu_r, b_r)^T$

    double precision                          :: drytol,&         !< dry tolerance
                                                 g                !< gravity

    !output
    double precision, dimension( 3 )                :: o_eL,& !< net-updates for the left cell
                                                       o_eR  !< net-updates for the right cell
    double precision, dimension( 3 ) :: sw       !< wave speeds/eigenvalues (negative/positive: waves travelling to the left/right cell)

#if CONFIG_TSUNAMI_AUGMENTED_RIEMANN_EIGEN_COEFFICIENTS
    double precision, dimension( 3 ) :: o_eigenCoefficients  !< eigencoefficients computed in the riemann solution
#endif

    !local
    integer                                              :: mw,&   !< wave number occuring in the Riemann solution (1: left wave, 2: (if defined) corrector wave, 3: right wave)
                                                            i !< number of the quantity (h = 1, hu=2, hv=3)
    double precision, dimension(3, 3 )    :: fw         !< f-waves
    double precision, dimension(3)                       :: wall           !< boundary conditions for the quantities
    double precision                                     :: hL,&           !< height, left cell
                                                            hR,&           !< height, right cell
                                                            huL,&          !< edge-normal momentum, left cell
                                                            huR,&          !< edge-normal momentum, right cell
                                                            hvL,&          !< edge-parallel momentum, left cell
                                                            hvR,&          !< edge-parallel momentum, right cell
                                                            bL,&           !< bathymetry, left cell
                                                            bR,&           !< bathymetry, right cell
                                                            uL,&           !< particle velocity, left cell in physical x-direction
                                                            uR,&           !< particle velocity, right cell in physical x-direction
                                                            vL,&           !< particle velocity, left cell in physical y-direction
                                                            vR,&           !< particle velocity, right cell in physical x-direction
                                                            hstar,&        !< height of the middle state
                                                            hstartest,&    !< maximum height h_l, h^*, h_r
                                                            uhat,&         !< Roe average of particle velocity
                                                            chat,&         !< wave speed relative to the fluid
                                                            phiL,&         !< normal momentum flux, left cell
                                                            phiR           !< normal momentum flux, right cell
    double precision                                    ::  sL,&           !< characteristic speed, left wave
                                                            sR,&           !< characteristic speed, right wave
                                                            sRoe1,&        !< Roe speed, left wave
                                                            sRoe2,&        !< Roe speed, right wave
                                                            sE1,&          !< Einfeldt speed, left wave
                                                            sE2,&          !< Einfeldt speed, right wave
                                                            s1m,&          !< Rarefaction speed, left side
                                                            s2m            !< Rarefaction speed, right side
    logical                                             ::  rare1,&        !< boolean whether the left wave is a rarefaction wave
                                                            rare2          !< boolean whether the right wave is a rarefaction wave


  ! copy cell values to local variables
  hL  = i_variablesLeft(1)
  huL = i_variablesLeft(2)
  bL  = i_variablesLeft(3)

  hR  = i_variablesRight(1)
  huR = i_variablesRight(2)
  bR  = i_variablesRight(3)

  ! set parallel momentums to zero (TODO: limiter)
  hvL = 0.d0
  hvR = 0.d0

  !******************************************************************
  !* necessary (changed) part of the GeoClaw subroutine rpn - start *
  !******************************************************************

  !reset net updates
  o_eL  = 0.d0;
  o_eR = 0.d0;


  !Initialize Riemann problem for grid interface
  sw = 0.d0
  fw     = 0.d0

  !zero (small) negative values if they exist
  if (hL.lt.0.d0) then
    hL = 0.d0 !rest of left variables will be set later
  endif

  if (hR.lt.0.d0) then
    hR = 0.d0 !rest of right variables will be set later
  endif

  !skip problem if in a completely dry area
  if (hL.le.drytol.and.hR.le.drytol) then
#if CONFIG_TSUNAMI_AUGMENTED_RIEMANN_EIGEN_COEFFICIENTS
    o_eigenCoefficients = 0;
#endif
    return;
  endif

  !check for wet/dry boundary
  if (hR.gt.drytol) then
    uR=huR/hR
    vR=hvR/hR
    phiR = 0.5d0*g*hR**2 + huR**2/hR
  else
    hR = 0.d0
    huR = 0.d0
    hvR = 0.d0
    uR = 0.d0
    vR = 0.d0
    phiR = 0.d0
  endif

  if (hL.gt.drytol) then
    uL=huL/hL
    vL=hvL/hL
    phiL = 0.5d0*g*hL**2 + huL**2/hL
  else
    hL=0.d0
    huL=0.d0
    hvL=0.d0
    uL=0.d0
    vL=0.d0
    phiL = 0.d0
  endif

  !per default there is no wall
  wall = 1.d0
  if (hR.le.drytol) then
    call riemanntype(hL,hL,uL,-uL,hstar,s1m,s2m,rare1,rare2,1,drytol,g)
    hstartest=max(hL,hstar)
    if (hstartest+bL.lt.bR) then !right state should become ghost values that mirror left for wall problem
      wall(2)=0.d0
      wall(3)=0.d0
      hR=hL
      huR=-huL
      bR=bL
      phiR=phiL
      uR=-uL
      vR=vL
    elseif (hL+bL.lt.bR) then
      bR=hL+bL
    endif
  elseif (hL.le.drytol) then ! right surface is lower than left topo
    call riemanntype(hR,hR,-uR,uR,hstar,s1m,s2m,rare1,rare2,1,drytol,g)
    hstartest=max(hR,hstar)
    if (hstartest+bR.lt.bL) then  !left state should become ghost values that mirror right
      wall(1)=0.d0
      wall(2)=0.d0
      hL=hR
      huL=-huR
      bL=bR
      phiL=phiR
      uL=-uR
      vL=vR
    elseif (hR+bR.lt.bL) then
      bL=hR+bR
    endif
  endif

  !determine wave speeds
  sL=uL-sqrt(g*hL) ! 1 wave speed of left state
  sR=uR+sqrt(g*hR) ! 2 wave speed of right state

  uhat=(sqrt(g*hL)*uL + sqrt(g*hR)*uR)/(sqrt(g*hR)+sqrt(g*hL)) ! Roe average
  chat=sqrt(g*0.5d0*(hR+hL)) ! Roe average
  sRoe1=uhat-chat ! Roe wave speed 1 wave
  sRoe2=uhat+chat ! Roe wave speed 2 wave

  sE1 = min(sL,sRoe1) ! Eindfeldt speed 1 wave
  sE2 = max(sR,sRoe2) ! Eindfeldt speed 2 wave

  !*******************
  !* call the solver *
  !*******************

  call riemann_aug_JCP(maxiter,3, 3, hL,hR,huL,huR,hvL,hvR,bL,bR,uL,&
       uR,vL,vR,phiL,phiR,sE1,sE2,drytol,g,sw,&
#if CONFIG_TSUNAMI_AUGMENTED_RIEMANN_EIGEN_COEFFICIENTS
     o_eigenCoefficients, &
#endif
      fw)


  !eliminate ghost fluxes for wall
  do mw=1,3
    sw(mw)=sw(mw)*wall(mw)
    do i=1,3
      fw(i,mw)=fw(i,mw)*wall(mw)
    enddo
  enddo

  !compute net updates
  do i=1,3
    do  mw=1,3
      if (sw(mw).lt.0.d0) then
       o_eL(i)=o_eL(i) + fw(i,mw);
      elseif (sw(mw).gt.0.d0) then
       o_eR(i)=o_eR(i) + fw(i,mw);
      else
        o_eL(i)=o_eL(i) + .5d0*fw(i,mw);
        o_eR(i)=o_eR(i) + .5d0*fw(i,mw);
      endif
    enddo
  enddo

  !******************************************************
  !* necessary part of the GeoClaw subroutine rpn - end *
  !******************************************************

  !add second order correction terms
  !do mw=1,3
    ! TODO!
    !correctionL = SIGN(1.d0,waveSpeeds(mw))
  !end do

  end subroutine c_bind_riemann_aug_JCP
end module
